/* automake.cc - -*- C++ -*- classes to work with Makefile.am
 * Copyright (C) 2000 Malenko Dmitry <maldim@mail.ru>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  
 */

#include <stdio.h>
#include <algorithm>

#include "automake.h"

//Target
Target::Target(const string &str)
{
	int start,end;
		
	//find name
	start = str.find_first_not_of(" \t");
	end = str.find(' ',start+1);
	name.assign(str,start,end - start);

	//find values
	start = str.find_first_not_of(" =\t",end);

	while(start != str.npos){
		end = str.find(' ',start+1);
		values.push_back(string(str,start,end-start));

		start = str.find_first_not_of(" \t",end);
	};
}

Target::Target(const Target &target)
{
	name = target.name;
	
	values.clear();
	values = target.values;
}

string Target::get_name(void) const
{
	return name;
}

void Target::add_value(const string &value)
{
	values.push_back(value);
}

void Target::remove_value(const string &value)
{
	values.remove(value);
}

bool Target::has_value(const string &value) const
{
	return find(values.begin(),values.end(),value) != values.end();
}

int Target::values_count(void) const
{
	return values.size();
}

Target::value_iterator Target::get_value_iterator(void) const
{
	return values.begin();
}

bool Target::iterator_last(Target::value_iterator iter) const
{
	return iter == values.end();
}

//Makefile
Makefile::Makefile(const string &_dir)
{
	dir = _dir;

	targets.clear();
	scripts.clear();

	load();
}

void Makefile::load(void)
{
	string filename = dir + "Makefile.am";
	char buf[128];
	string target;
	bool in_target = false, in_script = false;

	FILE *f = fopen(filename.c_str(),"r");

	if(!f)
		throw EIOError("Open Makefile.am for load failed.",filename);
	
	while(fgets(buf,128,f)){
		if(buf[0] == '#')
			continue;	//skip comments
	
		string foo = buf;
		
		if(!(in_target || in_script)){
			if(foo[foo.find_first_of("=:")] == '=')
				in_target = true;
			if(foo[foo.find_first_of("=:")] == ':')
				in_script = true;
		};

		if(in_target){
			if(foo[foo.find_last_not_of(" \t\n")] != '\\'){
				if(strrchr(buf,'\n'))
					*strrchr(buf,'\n') = '\0';
				target += buf;
				
				Target targ(target);	
				targets.push_back(targ);
				
				target = "";
				in_target = false;
			}else{
				*strrchr(buf,'\\') = '\0';
				target += buf;
			};
			continue;
		};
		
		if(in_script){
			if(foo[foo.find_last_not_of(" \t\n")] != '\\' &&
			   foo[foo.find_last_not_of(" \t\n")] != ':'){
				if(strrchr(buf,'\n'))
					*strrchr(buf,'\n') = '\0';
				target += buf;
				
				scripts.push_back(target);
				
				target = "";
				in_script = false;
			}else{
				target += buf;
			};
		};
	};
	
	if(!target.empty()){
		if(in_target){
			Target targ(target);
			targets.push_back(targ);
		};
		if(in_script){
			scripts.push_back(target);
		};
	};

	fclose(f);
}

void Makefile::save(void) throw(EIOError)
{
	string filename = dir + "/Makefile.am";
	string buf;
	target_list::iterator iter = targets.begin();
	script_list::iterator it = scripts.begin();

	FILE *f = fopen(filename.c_str(),"w");

	if(!f)
		throw EIOError("Open Makefile.am for save failed.",filename);
	
	while(iter != targets.end()){
		Target target = *iter++;
		buf = target.get_name() + " = ";
		fprintf(f,"%s \\\n",buf.c_str());

		Target::value_iterator it = target.get_value_iterator();
		fprintf(f,"\t%s ",(*it++).c_str());
		while(!target.iterator_last(it)){
			fprintf(f,"\\\n\t%s ",(*it++).c_str());
		};
		fprintf(f,"\n\n");
	};

	while(it != scripts.end()){
		fprintf(f,"%s\n",(*it++).c_str());
	};

	fclose(f);
}

const Target &Makefile::find_target(const string &target)
{
	target_list::iterator iter = find(targets.begin(),targets.end(),target);
	
	if(iter != targets.end())
		return *iter;
	else
		throw ETargetNotFound(target);	
}

void Makefile::add_target(const string &target,const string &values)
{
	Target targ(target + " = " + values);

	targets.push_back(targ);
}

void Makefile::remove_target(const string &target)
{
	target_list::iterator iter = find(targets.begin(),targets.end(),target);

	if(iter != targets.end())
		targets.erase(iter);
}

void Makefile::add_to_target(const string &target,const string &value)
{
	target_list::iterator iter = find(targets.begin(),targets.end(),target);

	if(iter != targets.end())
		(*iter).add_value(value);
}

void Makefile::remove_from_target(const string &target,const string &value)
{
	target_list::iterator iter = find(targets.begin(),targets.end(),target);

	if(iter != targets.end())
		(*iter).remove_value(value);
}

Makefile::target_iterator Makefile::get_target_iterator(void)
{
	return targets.begin();
}

bool Makefile::iterator_last(target_iterator iter)
{
	return iter == targets.end();
}
