/* projects.cc- Project classes -*- C++ -*-
 * Copyright (C) 2000 Malenko Dmitry <maldim@mail.ru>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  
 */

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

#include "conffile.h"
#include "util.h"
#include "automake.h"

#include "projects.h"

//"New file" dialogs
class NewCFileDialog : public DruidDialog
{
private:
	virtual void connect_signals(void)
	{
		DruidDialog::connect_signals();
		
		gtk_signal_connect(GTK_OBJECT(get_widget("nameentry")),"changed",GTK_SIGNAL_FUNC(on_name_changed),this);
	};
	
	//callbacks
	static void on_name_changed(GtkEntry *entry,NewCFileDialog *dia)
	{
		if(!strlen(gtk_entry_get_text(entry))){
			gnome_druid_set_buttons_sensitive(GNOME_DRUID(dia->get_widget("druid")),true,false,true);
		}else{
			gnome_druid_set_buttons_sensitive(GNOME_DRUID(dia->get_widget("druid")),true,true,true);
		};
	};
public:
	NewCFileDialog(void) :
		DruidDialog("titano","newcfiledia")
	{ };

	string get_name(void)
	{
		return gtk_entry_get_text(GTK_ENTRY(get_widget("nameentry")));
	};
	string get_ext(void)
	{
		return GTK_LABEL(GTK_OPTION_MENU(get_widget("extmenu"))->button.child)->label;
	};
	string get_desc(void)
	{
		return gtk_entry_get_text(GTK_ENTRY(get_widget("descentry")));
	};
	char get_type(void)
	{
		char res = 'b';

		if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(get_widget("s"))))
			res = 's';
		if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(get_widget("h"))))
			res = 'h';

		return res;			
	};
	bool use_template(void)
	{
		return gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(get_widget("templchbtn")));
	};
};


//Projects
Project::~Project()
{
	files_map::iterator iter = file_map.begin();

	while(iter != file_map.end()){
		delete (*iter++).second;
	};

	gtk_widget_destroy(GTK_WIDGET(notebook));
}

FileType Project::get_file_type(const string &name)
{
	FileType res = UNKNOWN;
	int pos = name.rfind('.');
	
	if(pos != name.npos){
		if(!name.compare(".h",pos,2))
			res = C_HEADER;
		if(!name.compare(".c",pos,2)||
		   !name.compare(".C",pos,2)||
		   !name.compare(".cc",pos,3)||
		   !name.compare(".cpp",pos,4)||
		   !name.compare(".cxx",pos,4))
			res = C_SOURCE;
	};

	return res;
}

void Project::load(void)
{
	ConfFile proj(filename);
	ConfFile::iterator iter;
	string file,open;
	bool modi = false;

	proj.read("General/name",name);
	proj.read("General/dir",dir);
	chdir(dir.c_str());
	proj.read("General/version",version);

	iter = proj.get_iterator("Files");
	iter = proj.iterator_next(iter,file,open);
	while(iter){
		if(file_exists(file)){
			file_list.push_back(file);
			create_item(file);

			if(open == "true")
				open_file(file);
		}else{
			if(!parent.question_yes_no("File " + file + " does not exist. Create?")){
				file_list.push_back(file);
				create_item(file);
				open_file(file);
			};
			modi = true;
		};

		iter = proj.iterator_next(iter,file,open);
	};
	modified = modi;
}

void Project::save(void)
{
	ConfFile proj(filename);
	files_list::iterator iter;

	proj.write("General/rev",1);

	proj.write("General/name",name);
	proj.write("General/dir",dir);
	proj.write("General/version",version);
	
	iter = file_list.begin();
	proj.clear_section("Files");
	while(iter != file_list.end()){
		proj.write("Files/" + *iter,file_is_opened(*iter));
		iter++;
	};

	proj.drop();

	modified = false;
}

void Project::save_as(const string &newname)
{
	string oldname = name;
	string olddir = dir;
	string oldfile = filename;
	
	int pos = newname.rfind('/');

	dir.assign(newname,0,pos);
	name.assign(newname,pos+1,newname.size()-pos-7);
	filename = newname;

	try{
		save();
	}catch(...){
		name = oldname;
		dir = olddir;
		filename = oldfile;
		throw;
	};

	parent.set_title(" - [C/C++] " + name);

	modified = false;
}

void Project::add_file(const string &fullname)
{
	file_list.push_back(fullname);
	create_item(fullname);

	add_am_file(fullname);

	modified = true;
}

void Project::remove_file(const string &fullname)
{
	if(file_is_opened(fullname))
		close_file(fullname);

	file_list.remove(fullname);
	remove_item(fullname);
	
	remove_am_file(fullname);

	modified = true;
}

void Project::open_file(const string &name)
{
	if(file_exists(name)){
		View *view = parent.get_file_view();

		cur_file = new File(name,view,notebook,NULL);
		file_map[name] = cur_file;

		modified = true;
	};
}

void Project::close_file(void)
{
	close_file(cur_file->get_fullname());
}

void Project::close_file(const string &name)
{
	File *file = file_map[name];
	delete file;
	
	file_map.erase(name);

	modified = true;
}

void Project::save_file_as(const string &name)
{
	string oldname = cur_file->get_fullname();
	cur_file->save_as(name);

	file_list.remove(oldname);
	file_list.push_back(name);

	file_map.erase(oldname);
	file_map[name] = cur_file;
	
	modified = true;
}

void Project::save_all_files(void)
{
	File *file;
	
	files_map::iterator iter = file_map.begin();

	while(iter != file_map.end()){
		file = (*iter++).second;
		if(file->modified())
			file->save();
	};
}

File *Project::operator[](const string &name)
{
	if(file_is_opened(name)){
		return file_map[name];
	}else{
		return NULL;
	};
}

bool Project::file_in_project(const string &name)
{
	return find(file_list.begin(),file_list.end(),name) != file_list.end();
}

bool Project::file_is_opened(const string &name)
{
	return file_map.count(name);
}

//Shell routines
void Project::autogen(void)
{
	chdir(dir.c_str());
	parent.execute("./autogen.sh",true,true,GTK_CLIST(parent.get_widget("compile_list")),GTK_CLIST(parent.get_widget("compile_list")));
}

void Project::configure(void)
{
	chdir(dir.c_str());
	parent.execute("./configure",true,true,GTK_CLIST(parent.get_widget("compile_list")),GTK_CLIST(parent.get_widget("compile_list")));
}

void Project::make(void)
{
	chdir(dir.c_str());
	parent.execute("make",true,true,GTK_CLIST(parent.get_widget("compile_list")),GTK_CLIST(parent.get_widget("compile_list")));
}

void Project::make_dist(void)
{
	chdir(dir.c_str());
	parent.execute("make dist",true,true,GTK_CLIST(parent.get_widget("compile_list")),GTK_CLIST(parent.get_widget("compile_list")));
}

//Automake routines
void Project::add_am_file(const string &file)
{
	int pos = file.rfind('/');
	string path(file,0,pos+1);
	string fname(file,pos+1,file.size() - pos - 1);
	
	Makefile am(path);

	am.add_to_target(name + "_SOURCES",fname);
	am.save();
}

void Project::remove_am_file(const string &file)
{
	int pos = file.rfind('/');
	string path(file,0,pos+1);
	string fname(file,pos+1,file.size() - pos - 1);
	
	Makefile am(path);

	am.remove_from_target(name + "_SOURCES",fname);
	am.save();
}

void Project::create_item(const string &name)
{
	GtkWidget *item = gtk_tree_item_new_with_label(adjust_file_name(name).c_str());
	gtk_object_set_data(GTK_OBJECT(item),"name",g_strdup(name.c_str()));
	gtk_signal_connect(GTK_OBJECT(item),"select",GTK_SIGNAL_FUNC(proj_item_selected),this);
	gtk_signal_connect(GTK_OBJECT(item),"deselect",GTK_SIGNAL_FUNC(proj_item_selected),this);
	gtk_widget_show(item);
	gtk_tree_append(get_file_tree(name),item);
	gtk_tree_item_expand(get_file_tree_item(name));
}

void Project::remove_item(const string &name)
{
	GtkTree *tree = get_file_tree(name);
	GList *l = tree->children;
	while(l && (char *)gtk_object_get_data(GTK_OBJECT(l->data),"name") != name)
		l = l->next;
	
	if(!l)
		return;
		
	GtkWidget *treeitem = GTK_WIDGET(l->data);
	gtk_widget_ref(GTK_WIDGET(tree));
	gtk_container_remove(GTK_CONTAINER(tree),treeitem);
	if(!GTK_WIDGET(tree)->parent)
		gtk_tree_item_set_subtree(get_file_tree_item(name),GTK_WIDGET(tree));
	else
		gtk_widget_unref(GTK_WIDGET(tree));
	gtk_tree_item_expand(get_file_tree_item(name));
}

string Project::adjust_file_name(const string &fullname)
{
	int len = dir.length() + 5;  //assuming all files under dir/src/
	string res(fullname,len,fullname.length() - len);
	return res;
}

//callbacks
void Project::on_file_switch(GtkNotebook *nb,gpointer data,gint n,Project *project)
{
	if(n != -1){
		GtkWidget *page = gtk_notebook_get_nth_page(nb,n);
		project->cur_file = (File *)gtk_object_get_data(GTK_OBJECT(page),"file");
	};
}

void Project::proj_item_selected(GtkItem *item,Project *project)
{
	string name = (char *)gtk_object_get_data(GTK_OBJECT(item),"name");
	
	if(!project->file_is_opened(name))
		project->open_file(name);
	else
		(*project)[name]->activate();
	
	project->parent.up_interface();
}

//CppProject
CppProject::~CppProject()
{
	gtk_widget_destroy(GTK_WIDGET(citem));
	gtk_widget_destroy(GTK_WIDGET(hitem));
	gtk_widget_destroy(GTK_WIDGET(oitem));
}

void CppProject::new_file(void)
{
	NewCFileDialog dia;

	try{
		if(dia.run()){
			string descr = dia.get_desc();
			
			switch(dia.get_type()){
			case 'b':
			case 'h':
				{
					string filename = dir + "/src/" + dia.get_name() + ".h";
									
					if(file_exists(filename)){
						if(parent.question_yes_no("Project " + filename + " already exists. Overwrite?"))
							goto cont;
					};

					if(dia.use_template()){
						parent.create_file(filename,parent.get_templdir() + "h_template",descr);
					}else{
						file_create(filename);
					};

					add_file(filename);
					open_file(filename);
				};

			cont:
				if(dia.get_type() == 'h')
					break;
			case 's':
				string filename = dir + "/src/" + dia.get_name() + dia.get_ext();
									
				if(file_exists(filename)){
					if(parent.question_yes_no("Project " + filename + " already exists. Overwrite?"))
						return;
				};

				if(dia.use_template()){
					parent.create_file(filename,parent.get_templdir() + "c_template",descr);
				}else{
					file_create(filename);
				};

				add_file(filename);
				open_file(filename);
				break;
			};
		};
	}catch(Exception e){
		parent.error(e);
	};
}

GtkTree *CppProject::get_file_tree(const string &name)
{
	switch(get_file_type(name)){
	case UNKNOWN:
		if(!otree){
				otree = GTK_TREE(gtk_tree_new());
				gtk_tree_item_set_subtree(oitem,GTK_WIDGET(otree));
		};
		return otree;
	case C_HEADER:
		if(!htree){
				htree = GTK_TREE(gtk_tree_new());
				gtk_tree_item_set_subtree(hitem,GTK_WIDGET(htree));
		};
		return htree;
	case C_SOURCE:
		if(!ctree){
				ctree = GTK_TREE(gtk_tree_new());
				gtk_tree_item_set_subtree(citem,GTK_WIDGET(ctree));
		};
		return ctree;
	};
}

GtkTreeItem *CppProject::get_file_tree_item(const string &name)
{
	switch(get_file_type(name)){
	case UNKNOWN:return oitem;
	case C_HEADER:return hitem;
	case C_SOURCE:return citem;
	};
}
