/* views.h - File view classes -*- C++ -*-
 * Copyright (C) 2000 Malenko Dmitry <maldim@mail.ru>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  
 */
 
#ifndef _VIEWS_H
#define _VIEWS_H

#include <string>

#include "iexceptions.h"

#include "gtkeditor/gtkeditor.h"

#define GTK

#include "scintilla/Platform.h"
#include "scintilla/ScintillaWidget.h"
#include "scintilla/Scintilla.h"

struct _ViewSettings
{
	bool spaces;       //use spaces instead of tabs
	int tab_stop;      //tabstop value
	bool viewws;       //spaces visible
	bool vieweol;      //end of lines vsisble
	bool indentation;  //view indentations guides
	string selcolor;   //selection color
	string selback;    //selection background color
};
typedef _ViewSettings ViewSettings;

class View
{
protected:
	virtual void setup(const string &file) = 0;
	string get_language(const string &file);
public:
	virtual GtkWidget *get_widget(void)
	{
		throw ENotSupported("Get view widget.","Not implemented.");
	};
	virtual char *get_buffer(void) const
	{
		throw ENotSupported("Get view buffer.","Not implemented.");
	};
	
	virtual bool modified(void) const
	{
		throw ENotSupported("Get modification state.","Not implemented.");
	};

	//settings
	virtual void set_settings(const ViewSettings &set) const
	{
	};
	
	//files
	virtual void load(const string &name)
	{
		throw ENotSupported("Load file to view.","Not implemented.");
	};
	virtual void save(const string &name) const
	{
		throw ENotSupported("Save view to file.","Not implemented.");
	};

	//position
	virtual void get_pos(int &line,int &col) const
	{
		throw ENotSupported("Get view position (line/col).","Not implemented.");
	};
	virtual unsigned long get_pos(void) const
	{
		throw ENotSupported("Get view position (char).","Not implemented.");
	};
	virtual void set_pos(int line,int col) const
	{
		throw ENotSupported("Set view position (line/col).","Not implemented.");
	};
	virtual void set_pos(int pos) const
	{
		throw ENotSupported("Set view position (char).","Not implemented.");
	};
	virtual unsigned long get_char_count(void) const
	{
		throw ENotSupported("Set view position (char).","Not implemented.");
	};
	virtual int get_line_count(void) const
	{
		throw ENotSupported("Set view position (char).","Not implemented.");
	};

	virtual void clear(void) const
	{
		throw ENotSupported("Clear view.","Not implemented.");
	};

	//edit
	virtual void undo(void) const
	{
		throw ENotSupported("Undo operation.","Not implemented.");
	};

	virtual void redo(void) const
	{
		throw ENotSupported("Redo operation.","Not implemented.");
	};

	virtual void cut(void) const
	{
		throw ENotSupported("Cut selection.","Not implemented.");
	};

	virtual void copy(void) const
	{
		throw ENotSupported("Copy selection.","Not implemented.");
	};

	virtual void paste(void) const
	{
		throw ENotSupported("Paste from clipboard.","Not implemented.");
	};

	//insert
	virtual void insert(const string &text) const
	{
		throw ENotSupported("Insert at current position.","Not implemented.");
	};

	virtual void insert(const string &text,int pos) const
	{
		throw ENotSupported("Insert at (char) position.","Not implemented.");
	};

	virtual void insert(const string &text,int line,int ch) const
	{
		throw ENotSupported("Insert at (line/col) position.","Not implemented.");
	};

	//find capabilities
	virtual void find(const string &text) const
	{
		throw ENotSupported("Find string.","Not implemented.");
	};

	virtual void replace(const string &text,const string &newtext) const
	{
		throw ENotSupported("Replace string.","Not implemented.");
	};
	
};

class GtkEditorView : public View
{
private:
	GtkEditor *editor;
	GtkTextView *view;
	GtkTextBuffer *buffer;
	GtkWidget *window;

	static GtkEditorScanner* cscanner;

	virtual void setup(const string &file);
public:
	GtkEditorView();
	virtual ~GtkEditorView();

	virtual GtkWidget *get_widget(void);
	virtual char *get_buffer(void) const;

	virtual bool modified(void) const;

	virtual void load(const string &name);
	virtual void save(const string &name) const;
	
	virtual void get_pos(int &line,int &col) const;
	virtual unsigned long get_pos(void) const;
	virtual void set_pos(int line,int col) const;
	virtual void set_pos(int pos) const;
	virtual unsigned long get_char_count(void) const;
	virtual int get_line_count(void) const;

	virtual void clear(void) const;
};

class ScintillaView : public View
{
private:
	Window w_editor;
	SciFnDirect fn_editor;
	long ptr_editor;

	virtual void setup(const string &file);

	long get_sci_color(const string &triplet) const;
	
	long send_editor(unsigned int msg,unsigned long wparam = 0,long lparam = 0) const;
	long send_editor_string(unsigned int msg,unsigned long wparam,const char *s) const;
	void get_range(int start,int end,char *text) const;
public:
	ScintillaView();
	virtual ~ScintillaView();

	virtual GtkWidget *get_widget(void);
	virtual char *get_buffer(void) const;

	virtual bool modified(void) const;

	virtual void set_settings(const ViewSettings &set) const;

	virtual void load(const string &name);
	virtual void save(const string &name) const;
	
	virtual void get_pos(int &line,int &col) const;
	virtual unsigned long get_pos(void) const;
	virtual void set_pos(int line,int col) const;
	virtual void set_pos(int pos) const;
	virtual unsigned long get_char_count(void) const;
	virtual int get_line_count(void) const;

	virtual void clear(void) const;

	virtual void undo(void) const;
	virtual void redo(void) const;

	virtual void cut(void) const;
	virtual void copy(void) const;
	virtual void paste(void) const;
	
	virtual void insert(const string &text) const;
};

class BonoboView : public View
{
};

#endif
